@@ -6,11 +6,14 @@ import sys
 import warnings
 from collections import defaultdict
 from dataclasses import dataclass, field
+from inspect import signature
+from textwrap import dedent
 from timeit import default_timer as timer
 from typing import Any, Dict, List, Optional, Tuple, Type
 
 from omegaconf import DictConfig
 
+from hydra._internal.callbacks import Callbacks
 from hydra._internal.sources_registry import SourcesRegistry
 from hydra.core.config_loader import ConfigLoader
 from hydra.core.singleton import Singleton
@@ -20,7 +23,8 @@ from hydra.plugins.launcher import Launcher
 from hydra.plugins.plugin import Plugin
 from hydra.plugins.search_path_plugin import SearchPathPlugin
 from hydra.plugins.sweeper import Sweeper
-from hydra.types import TaskFunction
+from hydra.types import HydraContext, TaskFunction
+from hydra.utils import instantiate
 
 
 @dataclass
@@ -86,7 +90,7 @@ class Plugins(metaclass=Singleton):
             if classname not in self.class_name_to_class.keys():
                 raise RuntimeError(f"Unknown plugin class : '{classname}'")
             clazz = self.class_name_to_class[classname]
-            plugin = internal_utils._instantiate_class(clazz=clazz, config=config)
+            plugin = instantiate(config=config, _target_=clazz)
             assert isinstance(plugin, Plugin)
 
         except ImportError as e:
@@ -103,36 +107,101 @@ class Plugins(metaclass=Singleton):
             "hydra._internal.core_plugins."
         )
 
+    @staticmethod
+    def _setup_plugin(
+        plugin: Any,
+        task_function: TaskFunction,
+        config: DictConfig,
+        config_loader: Optional[ConfigLoader] = None,
+        hydra_context: Optional[HydraContext] = None,
+    ) -> Any:
+        """
+        With HydraContext introduced in #1581, we need to set up the plugins in a way
+        that's compatible with both Hydra 1.0 and Hydra 1.1 syntax.
+        This method should be deleted in the next major release.
+        """
+        assert isinstance(plugin, Sweeper) or isinstance(plugin, Launcher)
+        assert (
+            config_loader is not None or hydra_context is not None
+        ), "config_loader and hydra_context cannot both be None"
+
+        param_keys = signature(plugin.setup).parameters.keys()
+
+        if "hydra_context" not in param_keys:
+            # DEPRECATED: remove in 1.2
+            # hydra_context will be required in 1.2
+            warnings.warn(
+                message=dedent(
+                    """
+                    Plugin's setup() signature has changed in Hydra 1.1.
+                    Support for the old style will be removed in Hydra 1.2.
+                    For more info, check https://github.com/facebookresearch/hydra/pull/1581."""
+                ),
+                category=UserWarning,
+            )
+            config_loader = (
+                config_loader
+                if config_loader is not None
+                else hydra_context.config_loader  # type: ignore
+            )
+            plugin.setup(  # type: ignore
+                config=config,
+                config_loader=config_loader,
+                task_function=task_function,
+            )
+        else:
+            if hydra_context is None:
+                # hydra_context could be None when an incompatible Sweeper instantiates a compatible Launcher
+                assert config_loader is not None
+                hydra_context = HydraContext(
+                    config_loader=config_loader, callbacks=Callbacks()
+                )
+            plugin.setup(
+                config=config, hydra_context=hydra_context, task_function=task_function
+            )
+        return plugin
+
     def instantiate_sweeper(
         self,
-        config: DictConfig,
-        config_loader: ConfigLoader,
+        *,
+        hydra_context: HydraContext,
         task_function: TaskFunction,
+        config: DictConfig,
     ) -> Sweeper:
         Plugins.check_usage(self)
         if config.hydra.sweeper is None:
             raise RuntimeError("Hydra sweeper is not configured")
         sweeper = self._instantiate(config.hydra.sweeper)
-        assert isinstance(sweeper, Sweeper)
-        sweeper.setup(
-            config=config, config_loader=config_loader, task_function=task_function
+        sweeper = self._setup_plugin(
+            plugin=sweeper,
+            task_function=task_function,
+            config=config,
+            config_loader=None,
+            hydra_context=hydra_context,
         )
+        assert isinstance(sweeper, Sweeper)
         return sweeper
 
     def instantiate_launcher(
         self,
-        config: DictConfig,
-        config_loader: ConfigLoader,
         task_function: TaskFunction,
+        config: DictConfig,
+        config_loader: Optional[ConfigLoader] = None,
+        hydra_context: Optional[HydraContext] = None,
     ) -> Launcher:
         Plugins.check_usage(self)
         if config.hydra.launcher is None:
             raise RuntimeError("Hydra launcher is not configured")
+
         launcher = self._instantiate(config.hydra.launcher)
-        assert isinstance(launcher, Launcher)
-        launcher.setup(
-            config=config, config_loader=config_loader, task_function=task_function
+        launcher = self._setup_plugin(
+            plugin=launcher,
+            config=config,
+            task_function=task_function,
+            config_loader=config_loader,
+            hydra_context=hydra_context,
         )
+        assert isinstance(launcher, Launcher)
         return launcher
 
     @staticmethod
@@ -175,7 +244,7 @@ class Plugins(metaclass=Singleton):
                         )
                         for w in recorded_warnings:
                             warnings.showwarning(
-                                message=w.message,  # type: ignore
+                                message=w.message,
                                 category=w.category,
                                 filename=w.filename,
                                 lineno=w.lineno,
