@@ -7,16 +7,21 @@ import sys
 import warnings
 from contextlib import contextmanager
 from dataclasses import dataclass
-from os.path import basename, dirname, splitext
+from datetime import datetime
+from enum import Enum
+from os.path import splitext
 from pathlib import Path
-from time import localtime, strftime
-from typing import Any, Dict, Optional, Sequence, Tuple, Union, cast
+from textwrap import dedent
+from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Union, cast
 
 from omegaconf import DictConfig, OmegaConf, open_dict, read_write
 
 from hydra.core.hydra_config import HydraConfig
 from hydra.core.singleton import Singleton
-from hydra.types import TaskFunction
+from hydra.types import HydraContext, TaskFunction
+
+if TYPE_CHECKING:
+    from hydra._internal.callbacks import Callbacks
 
 log = logging.getLogger(__name__)
 
@@ -31,14 +36,16 @@ def simple_stdout_log_config(level: int = logging.INFO) -> None:
 
 
 def configure_log(
-    log_config: DictConfig, verbose_config: Union[bool, str, Sequence[str]]
+    log_config: DictConfig,
+    verbose_config: Union[bool, str, Sequence[str]] = False,
 ) -> None:
     assert isinstance(verbose_config, (bool, str)) or OmegaConf.is_list(verbose_config)
     if log_config is not None:
         conf: Dict[str, Any] = OmegaConf.to_container(  # type: ignore
             log_config, resolve=True
         )
-        logging.config.dictConfig(conf)
+        if conf["root"] is not None:
+            logging.config.dictConfig(conf)
     else:
         # default logging to stdout
         root = logging.getLogger()
@@ -66,7 +73,7 @@ def configure_log(
 
 def _save_config(cfg: DictConfig, filename: str, output_dir: Path) -> None:
     output_dir.mkdir(parents=True, exist_ok=True)
-    with open(str(output_dir / filename), "w") as file:
+    with open(str(output_dir / filename), "w", encoding="utf-8") as file:
         file.write(OmegaConf.to_yaml(cfg))
 
 
@@ -78,13 +85,37 @@ def filter_overrides(overrides: Sequence[str]) -> Sequence[str]:
     return [x for x in overrides if not x.startswith("hydra.")]
 
 
+def _get_callbacks_for_run_job(hydra_context: Optional[HydraContext]) -> "Callbacks":
+    if hydra_context is None:
+        # DEPRECATED: remove in 1.2
+        # hydra_context will be required in 1.2
+        warnings.warn(
+            message=dedent(
+                """
+                run_job's signature has changed in Hydra 1.1. Please pass in hydra_context.
+                Support for the old style will be removed in Hydra 1.2.
+                For more info, check https://github.com/facebookresearch/hydra/pull/1581."""
+            ),
+        )
+        from hydra._internal.callbacks import Callbacks
+
+        callbacks = Callbacks()
+    else:
+        callbacks = hydra_context.callbacks
+
+    return callbacks
+
+
 def run_job(
-    config: DictConfig,
     task_function: TaskFunction,
+    config: DictConfig,
     job_dir_key: str,
     job_subdir_key: Optional[str],
     configure_logging: bool = True,
+    hydra_context: Optional[HydraContext] = None,
 ) -> "JobReturn":
+    callbacks = _get_callbacks_for_run_job(hydra_context)
+
     old_cwd = os.getcwd()
     orig_hydra_cfg = HydraConfig.instance().cfg
     HydraConfig.instance().set_config(config)
@@ -124,11 +155,20 @@ def run_job(
             _save_config(config.hydra.overrides.task, "overrides.yaml", hydra_output)
 
         with env_override(hydra_cfg.hydra.job.env_set):
-            ret.return_value = task_function(task_cfg)
+            callbacks.on_job_start(config=config)
+            try:
+                ret.return_value = task_function(task_cfg)
+                ret.status = JobStatus.COMPLETED
+            except Exception as e:
+                ret.return_value = e
+                ret.status = JobStatus.FAILED
+
         ret.task_name = JobRuntime.instance().get("name")
 
         _flush_loggers()
 
+        callbacks.on_job_end(config=config, job_return=ret)
+
         return ret
     finally:
         HydraConfig.instance().cfg = orig_hydra_cfg
@@ -141,18 +181,17 @@ def get_valid_filename(s: str) -> str:
 
 
 def setup_globals() -> None:
-    def register(name: str, f: Any) -> None:
-        try:
-            OmegaConf.register_resolver(name, f)
-        except AssertionError:
-            # calling it again in no_workers mode will throw. safe to ignore.
-            pass
-
     # please add documentation when you add a new resolver
-    register("now", lambda pattern: strftime(pattern, localtime()))
-    register(
+    OmegaConf.register_new_resolver(
+        "now",
+        lambda pattern: datetime.now().strftime(pattern),
+        use_cache=True,
+        replace=True,
+    )
+    OmegaConf.register_new_resolver(
         "hydra",
         lambda path: OmegaConf.select(cast(DictConfig, HydraConfig.get()), path),
+        replace=True,
     )
 
     vi = sys.version_info
@@ -161,17 +200,41 @@ def setup_globals() -> None:
         "minor": f"{vi[0]}.{vi[1]}",
         "micro": f"{vi[0]}.{vi[1]}.{vi[2]}",
     }
-    register("python_version", lambda level="minor": version_dict.get(level))
+    OmegaConf.register_new_resolver(
+        "python_version", lambda level="minor": version_dict.get(level), replace=True
+    )
+
+
+class JobStatus(Enum):
+    UNKNOWN = 0
+    COMPLETED = 1
+    FAILED = 2
 
 
 @dataclass
 class JobReturn:
     overrides: Optional[Sequence[str]] = None
-    return_value: Any = None
     cfg: Optional[DictConfig] = None
     hydra_cfg: Optional[DictConfig] = None
     working_dir: Optional[str] = None
     task_name: Optional[str] = None
+    status: JobStatus = JobStatus.UNKNOWN
+    _return_value: Any = None
+
+    @property
+    def return_value(self) -> Any:
+        assert self.status != JobStatus.UNKNOWN, "return_value not yet available"
+        if self.status == JobStatus.COMPLETED:
+            return self._return_value
+        else:
+            sys.stderr.write(
+                f"Error executing job with overrides: {self.overrides}" + os.linesep
+            )
+            raise self._return_value
+
+    @return_value.setter
+    def return_value(self, value: Any) -> None:
+        self._return_value = value
 
 
 class JobRuntime(metaclass=Singleton):
@@ -190,40 +253,17 @@ class JobRuntime(metaclass=Singleton):
         self.conf[key] = value
 
 
-def split_config_path(
-    config_path: Optional[str], config_name: Optional[str]
-) -> Tuple[Optional[str], Optional[str]]:
-    if config_path is None or config_path == "":
-        return None, config_name
-    split_file = splitext(config_path)
-    if split_file[1] in (".yaml", ".yml"):
-        # assuming dir/config.yaml form
-        config_file: Optional[str] = basename(config_path)
-        config_dir: Optional[str] = dirname(config_path)
-        msg = (
-            "\nUsing config_path to specify the config name is deprecated, specify the config name via config_name"
-            "\nSee https://hydra.cc/docs/next/upgrades/0.11_to_1.0/config_path_changes"
-        )
-        warnings.warn(category=UserWarning, message=msg)
-    else:
-        # assuming dir form without a config file.
-        config_file = None
-        config_dir = config_path
-
-    if config_dir == "":
-        config_dir = None
-
-    if config_file == "":
-        config_file = None
-
-    if config_file is not None:
-        if config_name is not None:
-            raise ValueError(
-                "Config name should be specified in either normalized_config_path or config_name, but not both"
+def validate_config_path(config_path: Optional[str]) -> None:
+    if config_path is not None:
+        split_file = splitext(config_path)
+        if split_file[1] in (".yaml", ".yml"):
+            msg = dedent(
+                """\
+            Using config_path to specify the config name is not supported, specify the config name via config_name.
+            See https://hydra.cc/docs/next/upgrades/0.11_to_1.0/config_path_changes
+            """
             )
-        config_name = config_file
-
-    return config_dir, config_name
+            raise ValueError(msg)
 
 
 @contextmanager
